@@ -141,16 +141,21 @@ public:
         - SID_LIST: messages with session identifiers specified in the set.
         - UID_LIST: messages with unique identifiers specified in the set.
         - SUBJECT: messages that contain the specified string in the subject.
+        - BODY: messages that contain the specified string in the body.
         - FROM: messages that contain the specified string in the 'from' field.
         - TO: messages that contain the specified string in the 'to' field.
         - BEFORE_DATE: messages whose internal date is earlier than the specified date.
         - ON_DATE: messages whose internal date is within the specified date.
         - SINCE_DATE: messages whose internal date is within or later than the specified date.
+        - NEW: messages that have the `\Recent` flag set but not `\Seen`.
+        - RECENT: messages that have the `\Recent` flag set.
+        - SEEN: messages that have the `\Seen` flag set.
+        - UNSEEN: messages that do not have the `\Seen` flag set.
 
-        @todo Criterion for the body.
         @todo Criteria for the flags set.
+        @todo Negative (not) criteria.
         **/
-        enum key_type {ALL, SID_LIST, UID_LIST, SUBJECT, FROM, TO, BEFORE_DATE, ON_DATE, SINCE_DATE} key;
+        enum key_type {ALL, SID_LIST, UID_LIST, SUBJECT, BODY, FROM, TO, BEFORE_DATE, ON_DATE, SINCE_DATE, NEW, RECENT, SEEN, UNSEEN} key;
 
         /**
         Condition value type to be used as message search criteria.
@@ -240,7 +245,7 @@ public:
     Selecting a mailbox.
 
     @param mailbox    Mailbox to select.
-    @param read_only  Flag if the selected mailbox is only readable of also writable.
+    @param read_only  Flag if the selected mailbox is only readable or also writable.
     @return           Mailbox statistics.
     @throw imap_error Selecting mailbox failure.
     @throw imap_error Parsing failure.
@@ -264,8 +269,27 @@ public:
     @throw *           `fetch(const list<messages_range_t>&, map<unsigned long, message>&, bool, bool, codec::line_len_policy_t)`.
     @todo              Add server error messages to exceptions.
     **/
+    [[deprecated]]
     void fetch(const std::string& mailbox, unsigned long message_no, message& msg, bool header_only = false);
 
+    /**
+    Fetching a message from the mailbox.
+
+    Some servers report success if a message with the given number does not exist, so the method returns with the empty `msg`. Other considers
+    fetching non-existing message to be an error, and an exception is thrown.
+
+    @param mailbox     Mailbox to fetch from.
+    @param message_no  Number of the message to fetch.
+    @param msg         Message to store the result.
+    @param is_uid      Using a message uid number instead of a message sequence number.
+    @param header_only Flag if only the message header should be fetched.
+    @throw imap_error  Fetching message failure.
+    @throw imap_error  Parsing failure.
+    @throw *           `fetch(const list<messages_range_t>&, map<unsigned long, message>&, bool, bool, codec::line_len_policy_t)`.
+    @todo              Add server error messages to exceptions.
+    **/
+    void fetch(const std::string& mailbox, unsigned long message_no, bool is_uid, message& msg, bool header_only = false);
+
     /**
     Fetching a message from an already selected mailbox.
 
@@ -392,6 +416,9 @@ public:
     /**
     Searching a mailbox.
 
+    The RFC 3501 section 6.4.4 does not specify whether another untagged response except the SEARCH can be obtained. However, there are IMAP servers which
+    send, for instance, the EXISTS response. Thus, such non-specified responses are ignored, instead of being reported as errors.
+
     @param conditions  List of conditions taken in conjuction way.
     @param results     Store resulting list of message sequence numbers or UIDs here.
                        Does not clear the list first, so that results can be accumulated.
@@ -520,6 +547,14 @@ protected:
     **/
     static std::string messages_range_list_to_string(std::list<messages_range_t> ranges);
 
+    /**
+    Escaping the double quote and backslashes.
+
+    @param text String to escape.
+    @return     Escaped string.
+    **/
+    static std::string to_astring(const std::string& text);
+
     /**
     Untagged response character as defined by the protocol.
     **/
@@ -954,7 +989,7 @@ protected:
 /**
 Error thrown by IMAP client.
 **/
-class imap_error : public std::runtime_error
+class imap_error : public dialog_error
 {
 public:
 
@@ -962,19 +997,27 @@ public:
     Calling parent constructor.
 
     @param msg  Error message.
+    @param details Detailed message.
     **/
-    explicit imap_error(const std::string& msg) : std::runtime_error(msg)
-    {
-    }
+    imap_error(const std::string& msg, const std::string& details);
 
     /**
     Calling parent constructor.
 
     @param msg  Error message.
+    @param details Detailed message.
     **/
-    explicit imap_error(const char* msg) : std::runtime_error(msg)
-    {
-    }
+    explicit imap_error(const char* msg, const std::string& details);
+
+    imap_error(const imap_error&) = default;
+
+    imap_error(imap_error&&) = default;
+
+    ~imap_error() = default;
+
+    imap_error& operator=(const imap_error&) = default;
+
+    imap_error& operator=(imap_error&&) = default;
 };
 
 
