@@ -36,6 +36,22 @@ copy at http://www.freebsd.org/copyright/freebsd-license.html.
 namespace mailio
 {
 
+/**
+Options to customize the formatting of a message. Used by message::format().
+**/
+struct message_format_options_t
+{
+    /**
+    Flag if the leading dot should be escaped.
+    **/
+    bool dot_escape = false;
+
+    /**
+    Flag whether bcc addresses should be added.
+    **/
+    bool add_bcc_header = false;
+};
+
 
 /**
 Mail message and applied parsing/formatting algorithms.
@@ -44,26 +60,6 @@ class MAILIO_EXPORT message : public mime
 {
 public:
 
-    /**
-    Enclosing character for the mail address begin.
-    **/
-    static const char ADDRESS_BEGIN_CHAR = '<';
-
-    /**
-    String representation of the mail address start enclosing character.
-    **/
-    static const std::string ADDRESS_BEGIN_STR;
-
-    /**
-    Enclosing character for the mail address end.
-    **/
-    static const char ADDRESS_END_CHAR = '>';
-
-    /**
-    String representation of the mail address end enclosing character.
-    **/
-    static const std::string ADDRESS_END_STR;
-
     /**
     Character to separate mail addresses in a list.
     **/
@@ -119,18 +115,18 @@ public:
     If a line contains leading dot, then it can be escaped as required by mail protocols.
 
     @param message_str Resulting message as string.
-    @param dot_escape  Flag if the leading dot should be escaped.
-    @throw *           `format_header()`, `format_content(bool)`, `mime::format(string&, bool)`.
+    @param opts        Options to customize formatting.
+    @throw *           `format_header(format_options)`, `format_content(bool)`, `mime::format(string&, bool)`.
     **/
-    void format(std::string& message_str, bool dot_escape = false) const;
+    void format(std::string& message_str, const message_format_options_t& opts = message_format_options_t{}) const;
 
     /**
-    Overload of `format(string&, bool)`.
+    Overload of `format(string&, const message_format_options&)`.
 
     Because of the way the u8string is comverted to string, it's more expensive when used with C++20.
     **/
 #if defined(__cpp_char8_t)
-    void format(std::u8string& message_str, bool dot_escape = false) const;
+    void format(std::u8string& message_str,  const message_format_options_t& = message_format_options_t{}) const;
 #endif
 
     /**
@@ -392,8 +388,9 @@ public:
     Setting the subject.
 
     @param mail_subject Subject to set.
+    @param sub_codec    Codec of the subject to use.
     */
-    void subject(const std::string& mail_subject);
+    void subject(const std::string& mail_subject, codec::codec_t sub_codec = codec::codec_t::ASCII);
 
     /**
     Setting the raw subject.
@@ -408,8 +405,9 @@ public:
     Setting the subject.
 
     @param mail_subject Subject to set.
+    @param sub_codec    Codec of the subject to use.
     */
-    void subject(const std::u8string& mail_subject);
+    void subject(const std::u8string& mail_subject, codec::codec_t sub_codec = codec::codec_t::ASCII);
 
     /**
     Setting the raw subject.
@@ -447,19 +445,6 @@ public:
     **/
     void date_time(const boost::local_time::local_date_time& mail_dt);
 
-    /**
-    Attaching a file with the given media type.
-
-    @param att_strm Stream to read the attachment.
-    @param att_name Attachment name to set.
-    @param type     Attachment media type to set.
-    @param subtype  Attachment media subtype to set.
-    @throw *        `mime::content_type(const content_type_t&)`, `mime::content_transfer_encoding(content_transfer_encoding_t)`,
-                    `mime::content_disposition(content_disposition_t)`.
-    **/
-    [[deprecated]]
-    void attach(const std::istream& att_strm, const std::string& att_name, media_type_t type, const std::string& subtype);
-
     /**
     Attaching a list of streams.
 
@@ -470,7 +455,7 @@ public:
     @throw *           `mime::content_type(const content_type_t&)`, `mime::content_transfer_encoding(content_transfer_encoding_t)`,
                        `mime::content_disposition(content_disposition_t)`.
     **/
-    void attach(const std::list<std::tuple<std::istream&, std::string, content_type_t>>& attachments);
+    void attach(const std::list<std::tuple<std::istream&, string_t, content_type_t>>& attachments);
 
     /**
     Getting the number of attachments.
@@ -486,8 +471,9 @@ public:
     @param att_strm      Stream to write the attachment.
     @param att_name      Name of the attachment.
     @throw message_error Bad attachment index.
+    @todo                The attachment name should be also `string_t`.
     **/
-    void attachment(std::size_t index, std::ostream& att_strm, std::string& att_name) const;
+    void attachment(std::size_t index, std::ostream& att_strm, string_t& att_name) const;
 
     /**
     Adding another header.
@@ -514,7 +500,7 @@ public:
 
     @return Message headers.
     **/
-    std::multimap<std::string, std::string> headers() const;
+    const headers_t& headers() const;
 
 protected:
 
@@ -583,7 +569,6 @@ protected:
     **/
     static const std::string DATE_HEADER;
 
-
     /**
     Disposition notification header name.
     **/
@@ -594,16 +579,6 @@ protected:
     **/
     static const std::string MIME_VERSION_HEADER;
 
-    /**
-    Regex for the message id in the strict mode.
-    **/
-    static const std::string MESSAGE_ID_REGEX;
-
-    /**
-    Regex for the message id in the non-strict mode.
-    **/
-    static const std::string MESSAGE_ID_REGEX_NS;
-
     /**
     Formatting the header to a string.
 
@@ -613,7 +588,7 @@ protected:
     @throw message_error No sender for multiple authors.
     @throw *             `mime::format_header()`.
     **/
-    virtual std::string format_header() const;
+    virtual std::string format_header(bool add_bcc_header) const;
 
     /**
     Parsing a header line for a specific header.
@@ -636,21 +611,36 @@ protected:
     @throw message_error Formatting failure of address list, bad group name.
     @throw *             `format_address(const string&, const string&)`.
     **/
-    std::string format_address_list(const mailboxes& mailbox_list) const;
+    std::string format_address_list(const mailboxes& mailbox_list, const std::string& header_name = "") const;
 
     /**
     Formatting a name and an address.
 
-    If the name is in ASCII or the header codec set to UTF8, then it is written in raw format. Otherwise, the encoding is performed.
+    If the name is in ASCII or the header codec set to UTF8, then it is written in raw format. Otherwise, the encoding is performed. The header folding is
+    performed if necessary.
 
     @param name          Mail name.
     @param address       Mail address.
+    @param header_name   Header name of the address header.
     @return              The mail name and address formatted.
     @throw message_error Formatting failure of name.
     @throw message_error Formatting failure of address.
-    @todo                Deal with the line policy.
     **/
-    std::string format_address(const string_t& name, const std::string& address) const;
+    std::string format_address(const string_t& name, const std::string& address, const std::string& header_name) const;
+
+    /**
+    Formatting the subject which can be ASCII or UTF-8.
+
+    @return Formatted subject.
+    **/
+    std::string format_subject() const;
+
+    /**
+    Formatting email date.
+
+    @return Date for the email format.
+    **/
+    std::string format_date() const;
 
     /**
     Parsing a string into vector of names and addresses.
@@ -675,23 +665,6 @@ protected:
     **/
     boost::local_time::local_date_time parse_date(const std::string& date_str) const;
 
-    /**
-    Formatting the vector of IDs.
-
-    @param ids Vector of IDs.
-    @return    String of IDs in the angle brackets.
-    **/
-    static std::string format_many_ids(const std::vector<std::string>& ids);
-
-    /**
-    Formatting the ID.
-
-    @param id ID to format.
-    @return   ID within the angle brackets.
-    **/
-    static std::string format_many_ids(const std::string& id);
-
-
     /**
     Splitting string with Q encoded fragments into separate strings.
 
@@ -700,34 +673,6 @@ protected:
     **/
     static std::vector<std::string> split_qc_string(const std::string& text);
 
-    /**
-    Folding a long header.
-
-    @param header_line   Line of a header to be split into multiple lines by folding.
-    @return              Header split into multiple lines.
-    @throw message_error Header folding failure.
-
-    **/
-    std::string fold_header_line(const std::string& header_line) const;
-
-
-    /**
-    Parsing a string of IDs into a vector.
-
-    @param ids           String of IDs within the angle brackets.
-    @return              Vector of IDs.
-    @throw message_error Parsing failure of the message ID.
-    **/
-    std::vector<std::string> parse_many_ids(const std::string& ids) const;
-
-    /**
-    Formatting the subject which can be ASCII or UTF-8.
-
-    @return Formatted subject.
-    @todo   Folding to be moved into `format_header()`?
-    **/
-    string_t format_subject() const;
-
     /**
     Parsing a subject which can be ASCII or UTF-8.
 
@@ -738,7 +683,8 @@ protected:
     @throw message_error Parsing failure of Q encoding.
     @throw *             `q_codec::decode(const string&)`.
     **/
-    std::tuple<std::string, std::string> parse_subject(const std::string& subject);
+    std::tuple<std::string, std::string, codec::codec_t>
+    parse_subject(const std::string& subject);
 
     /**
     Parsing a name part of a mail ASCII or UTF-8 encoded.
@@ -816,35 +762,11 @@ protected:
     /**
     Other headers not included into the known ones.
     **/
-    std::multimap<std::string, std::string> headers_;
+    headers_t headers_;
 };
 
-
-/**
-Exception reported by `message` class.
-**/
-class message_error : public std::runtime_error
-{
-public:
-
-    /**
-    Calling parent constructor.
-
-    @param msg Error message.
-    **/
-    explicit message_error(const std::string& msg) : std::runtime_error(msg)
-    {
-    }
-
-    /**
-    Calling parent constructor.
-
-    @param msg Error message.
-    **/
-    explicit message_error(const char* msg) : std::runtime_error(msg)
-    {
-    }
-};
+[[deprecated]]
+typedef mime_error message_error;
 
 
 } // namespace mailio
