@@ -81,6 +81,12 @@ string imap::messages_range_list_to_string(list<messages_range_t> ranges)
 }
 
 
+string imap::to_astring(const string& text)
+{
+    return codec::surround_string(codec::escape_string(text, "\"\\"));
+}
+
+
 imap::search_condition_t::search_condition_t(imap::search_condition_t::key_type condition_key, imap::search_condition_t::value_type condition_value) :
     key(condition_key), value(condition_value)
 {
@@ -108,6 +114,10 @@ imap::search_condition_t::search_condition_t(imap::search_condition_t::key_type
                 imap_string = "SUBJECT " + QUOTED_STRING_SEPARATOR + std::get<string>(value) + QUOTED_STRING_SEPARATOR;
                 break;
 
+            case BODY:
+                imap_string = "BODY " + QUOTED_STRING_SEPARATOR + std::get<string>(value) + QUOTED_STRING_SEPARATOR;
+                break;
+
             case FROM:
                 imap_string = "FROM " + QUOTED_STRING_SEPARATOR + std::get<string>(value) + QUOTED_STRING_SEPARATOR;
                 break;
@@ -128,13 +138,29 @@ imap::search_condition_t::search_condition_t(imap::search_condition_t::key_type
                 imap_string = "SINCE " + imap_date_to_string(std::get<boost::gregorian::date>(value));
                 break;
 
+            case NEW:
+                imap_string = "NEW";
+                break;
+
+            case RECENT:
+                imap_string = "RECENT";
+                break;
+
+            case SEEN:
+                imap_string = "SEEN";
+                break;
+
+            case UNSEEN:
+                imap_string = "UNSEEN";
+                break;
+
             default:
                 break;
         }
     }
-    catch (std::bad_variant_access&)
+    catch (std::bad_variant_access& exc)
     {
-        throw imap_error("Invaid search condition.");
+        throw imap_error("Invaid search condition.", exc.what());
     }
 }
 
@@ -209,9 +235,9 @@ auto imap::select(const string& mailbox, bool read_only) -> mailbox_stat_t
 {
     string cmd;
     if (read_only)
-        cmd = format("EXAMINE " + QUOTED_STRING_SEPARATOR + mailbox + QUOTED_STRING_SEPARATOR);
+        cmd = format("EXAMINE " + to_astring(mailbox));
     else
-        cmd = format("SELECT " + QUOTED_STRING_SEPARATOR + mailbox + QUOTED_STRING_SEPARATOR);
+        cmd = format("SELECT " + to_astring(mailbox));
     dlg_->send(cmd);
 
     mailbox_stat_t stat;
@@ -244,19 +270,19 @@ auto imap::select(const string& mailbox, bool read_only) -> mailbox_stat_t
                         if (iequals(key->atom, "UNSEEN"))
                         {
                             if (value->token_type != response_token_t::token_type_t::ATOM)
-                                throw imap_error("Parsing failure.");
+                                throw imap_error("Number expected for unseen.", "Line=`" + line + "`.");
                             stat.messages_first_unseen = stoul(value->atom);
                         }
                         else if (iequals(key->atom, "UIDNEXT"))
                         {
                             if (value->token_type != response_token_t::token_type_t::ATOM)
-                                throw imap_error("Parsing failure.");
+                                throw imap_error("Number expected for uidnext.", "Line=`" + line + "`.");
                             stat.uid_next = stoul(value->atom);
                         }
                         else if (iequals(key->atom, "UIDVALIDITY"))
                         {
                             if (value->token_type != response_token_t::token_type_t::ATOM)
-                                throw imap_error("Parsing failure.");
+                                throw imap_error("Number expected for uidvalidity.", "Line=`" + line + "`.");
                             stat.uid_validity = stoul(value->atom);
                         }
                     }
@@ -285,26 +311,26 @@ auto imap::select(const string& mailbox, bool read_only) -> mailbox_stat_t
             else if (parsed_line.tag == to_string(tag_))
             {
                 if (!parsed_line.result.has_value() || parsed_line.result.value() != tag_result_response_t::OK)
-                    throw imap_error("Select or examine mailbox failure.");
+                    throw imap_error("Select or examine mailbox failure.", "Response=`" + parsed_line.response + "`.");
 
                 has_more = false;
             }
             else
-                throw imap_error("Parsing failure.");
+                throw imap_error("Parsing failure.", "Line=`" + line + "`.");
         }
     }
-    catch (const invalid_argument&)
+    catch (const invalid_argument& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Integer expected.", exc.what());
     }
-    catch (const out_of_range&)
+    catch (const out_of_range& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Integer expected.", exc.what());
     }
 
     // The EXISTS and RECENT are required, the others may be missing in earlier protocol versions.
     if (!exists_found || !recent_found)
-        throw imap_error("Parsing failure.");
+        throw imap_error("No number of existing or recent messages.", "");
 
     reset_response_parser();
     return stat;
@@ -318,22 +344,39 @@ void imap::fetch(const string& mailbox, unsigned long message_no, message& msg,
 }
 
 
+void imap::fetch(const string& mailbox, unsigned long message_no, bool is_uid, message& msg, bool header_only)
+{
+    select(mailbox);
+    fetch(message_no, msg, is_uid, header_only);
+}
+
+
 void imap::fetch(unsigned long message_no, message& msg, bool is_uid, bool header_only)
 {
     list<messages_range_t> messages_range;
     messages_range.push_back(imap::messages_range_t(message_no, message_no));
     map<unsigned long, message> found_messages;
     fetch(messages_range, found_messages, is_uid, header_only, msg.line_policy());
-    msg = std::move(found_messages.begin()->second);
+    if (!found_messages.empty())
+        msg = std::move(found_messages.begin()->second);
 }
 
 
-// Fetching literal is the only place where line is ended with LF only, instead of CRLF. Thus, `receive(true)` and counting EOLs is performed.
+/*
+Although the RFC mandates MIME messages to have lines ending with CRLF, some email clients put LF only. For that reason, the method `dialog::receive(true)`
+is being used for fetching a literal. That way, the EOL counting is properly performed. This is probably the case for the strict mode policy.
+
+According to the RFC 3501 section 6.4.5, the untagged response of the fetch command is not mandatory. Thus, some servers return just the tagged response if
+no message is found.
+
+The fetch untagged responses provide messages. In the last tagged response the status is obtained. Then, string literals can be parsed to validate the MIME
+format.
+*/
 void imap::fetch(const list<messages_range_t>& messages_range, map<unsigned long, message>& found_messages, bool is_uids, bool header_only,
     codec::line_len_policy_t line_policy)
 {
     if (messages_range.empty())
-        throw imap_error("Empty messages range.");
+        throw imap_error("Empty messages range.", "");
 
     const string RFC822_TOKEN = string("RFC822") + (header_only ? ".HEADER" : "");
     const string message_ids = messages_range_list_to_string(messages_range);
@@ -344,8 +387,9 @@ void imap::fetch(const list<messages_range_t>& messages_range, map<unsigned long
     cmd.append("FETCH " + message_ids + TOKEN_SEPARATOR_STR + RFC822_TOKEN);
     dlg_->send(format(cmd));
 
-    // Stores messages as string literals for parsing after the OK response.
+    // Stores messages as strings indexed by the messages number.
     map<unsigned long, string> msg_str;
+    // Flag whether the response line is the last one i.e. the tagged response.
     bool has_more = true;
     try
     {
@@ -355,21 +399,23 @@ void imap::fetch(const list<messages_range_t>& messages_range, map<unsigned long
             string line = dlg_->receive();
             tag_result_response_t parsed_line = parse_tag_result(line);
 
+            // The untagged response collects all messages.
             if (parsed_line.tag == UNTAGGED_RESPONSE)
             {
                 parse_response(parsed_line.response);
 
                 if (mandatory_part_.front()->token_type != response_token_t::token_type_t::ATOM)
-                    throw imap_error("Fetching message failure.");
-                unsigned long msg_no = stoul(mandatory_part_.front()->atom);
+                    throw imap_error("Number expected when fetching a message.", "Response=`" + parsed_line.response + ".`");
+
+                unsigned long sequence_no = stoul(mandatory_part_.front()->atom);
                 mandatory_part_.pop_front();
-                if (msg_no == 0)
-                    throw imap_error("Fetching message failure.");
+                if (sequence_no == 0)
+                    throw imap_error("Zero sequence number when fetching a message.", "");
 
                 if (!iequals(mandatory_part_.front()->atom, "FETCH"))
-                    throw imap_error("Fetching message failure.");
+                    throw imap_error("Expecting the fetch atom.", "Response=`" + parsed_line.response + ".`");
 
-                unsigned long uid = 0;
+                unsigned long uid_no = 0;
                 shared_ptr<response_token_t> literal_token = nullptr;
                 for (auto part : mandatory_part_)
                     if (part->token_type == response_token_t::token_type_t::LIST)
@@ -380,14 +426,14 @@ void imap::fetch(const list<messages_range_t>& messages_range, map<unsigned long
                                 {
                                     token++;
                                     if (token == part->parenthesized_list.end())
-                                        throw imap_error("Parsing failure.");
-                                    uid = stoul((*token)->atom);
+                                        throw imap_error("No uid number when fetching a message.", "");
+                                    uid_no = stoul((*token)->atom);
                                 }
                                 else if (iequals((*token)->atom, RFC822_TOKEN))
                                 {
                                     token++;
                                     if (token == part->parenthesized_list.end() || (*token)->token_type != response_token_t::token_type_t::LITERAL)
-                                        throw imap_error("Parsing failure.");
+                                        throw imap_error("No literal when fetching a message.", "");
                                     literal_token = *token;
                                     break;
                                 }
@@ -409,19 +455,19 @@ void imap::fetch(const list<messages_range_t>& messages_range, map<unsigned long
                         string line = dlg_->receive(true);
                         if (!line.empty())
                             trim_eol(line);
+                        // There could be a parenthesized list after the string literal. Read it and do nothing.
                         parse_response(line);
                     }
-                    msg_str.emplace(is_uids ? uid : msg_no, move(literal_token->literal));
+                    msg_str.emplace(is_uids ? uid_no : sequence_no, move(literal_token->literal));
 
                     // If no UID was found, but we asked for them, it's an error.
-                    if (is_uids && uid == 0)
+                    if (is_uids && uid_no == 0)
                     {
-                        throw imap_error("Parsing failure.");
+                        throw imap_error("No UID when fetching a message.", "");
                     }
                 }
-                else
-                    throw imap_error("Parsing failure.");
             }
+            // The tagged response determines status and parses provided messages in `msg_str`.
             else if (parsed_line.tag == to_string(tag_))
             {
                 if (parsed_line.result.value() == tag_result_response_t::OK)
@@ -430,25 +476,25 @@ void imap::fetch(const list<messages_range_t>& messages_range, map<unsigned long
                     for (const auto& ms : msg_str)
                     {
                         message msg;
-                        msg.line_policy(line_policy, line_policy);
+                        msg.line_policy(line_policy);
                         msg.parse(ms.second);
                         found_messages.emplace(ms.first, move(msg));
                     }
                 }
                 else
-                    throw imap_error("Fetching message failure.");
+                    throw imap_error("Fetching message failure.", "Response=`" + parsed_line.response + "`.");
             }
             else
-                throw imap_error("Parsing failure.");
+                throw imap_error("Invalid tag when fetching a message.", "Parsed tag=`" + parsed_line.tag + "`.");
         }
     }
-    catch (const invalid_argument&)
+    catch (const invalid_argument& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
-    catch (const out_of_range&)
+    catch (const out_of_range& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
 
     reset_response_parser();
@@ -466,15 +512,15 @@ void imap::append(const list<string>& folder_name, const message& msg)
 void imap::append(const string& folder_name, const message& msg)
 {
     string msg_str;
-    msg.format(msg_str, true);
+    msg.format(msg_str, message_format_options_t{true, false});
 
-    string cmd = "APPEND " + folder_name;
+    string cmd = "APPEND " + to_astring(folder_name);
     cmd.append(" {" + to_string(msg_str.size()) + "}");
     dlg_->send(format(cmd));
     string line = dlg_->receive();
     tag_result_response_t parsed_line = parse_tag_result(line);
     if (parsed_line.result == tag_result_response_t::BAD || parsed_line.tag != CONTINUE_RESPONSE)
-        throw imap_error("Message appending failure.");
+        throw imap_error("Message appending failure.", "Response=`" + parsed_line.response + "`.");
 
     dlg_->send(msg_str);
     bool has_more = true;
@@ -485,11 +531,11 @@ void imap::append(const string& folder_name, const message& msg)
         if (parsed_line.tag == to_string(tag_))
         {
             if (parsed_line.result != tag_result_response_t::OK)
-                throw imap_error("Message appending failure.");
+                throw imap_error("Message appending failure.", "Line=`" + line + "`.");
             has_more = false;
         }
         else if (parsed_line.tag != UNTAGGED_RESPONSE)
-            throw imap_error("Message appending failure.");
+            throw imap_error("Expecting the untagged response.", "Tag=`" + parsed_line.tag + "`.");
     }
 }
 
@@ -499,7 +545,7 @@ auto imap::statistics(const string& mailbox, unsigned int info) -> mailbox_stat_
     // It doesn't like search terms it doesn't recognize.
     // Some older protocol versions or some servers may not support them.
     // So unseen uidnext and uidvalidity are optional.
-    string cmd = "STATUS " + QUOTED_STRING_SEPARATOR + mailbox + QUOTED_STRING_SEPARATOR + " (messages recent";
+    string cmd = "STATUS " + to_astring(mailbox) + " (messages recent";
     if (info & mailbox_stat_t::UNSEEN)
         cmd += " unseen";
     if (info & mailbox_stat_t::UID_NEXT)
@@ -524,7 +570,7 @@ auto imap::statistics(const string& mailbox, unsigned int info) -> mailbox_stat_
             {
                 parse_response(parsed_line.response);
                 if (!iequals(mandatory_part_.front()->atom, "STATUS"))
-                    throw imap_error("Getting statistics failure.");
+                    throw imap_error("Expecting the status atom.", "Line=`" + line + "`.");
                 mandatory_part_.pop_front();
 
                 bool mess_found = false, recent_found = false;
@@ -572,26 +618,26 @@ auto imap::statistics(const string& mailbox, unsigned int info) -> mailbox_stat_
                     }
                 // The MESSAGES and RECENT are required.
                 if (!mess_found || !recent_found)
-                    throw imap_error("Parsing failure.");
+                    throw imap_error("No messages or recent messages found.", "");
             }
             else if (parsed_line.tag == to_string(tag_))
             {
                 if (parsed_line.result.value() == tag_result_response_t::OK)
                     has_more = false;
                 else
-                    throw imap_error("Getting statistics failure.");
+                    throw imap_error("Getting statistics failure.", "Line=`" + line + "`.");
             }
             else
-                throw imap_error("Parsing failure.");
+                throw imap_error("Parsing failure.", "Tag=`" + parsed_line.tag + "`.");
         }
     }
-    catch (const invalid_argument&)
+    catch (const invalid_argument& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
-    catch (const out_of_range&)
+    catch (const out_of_range& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
 
     reset_response_parser();
@@ -644,24 +690,24 @@ void imap::remove(unsigned long message_no, bool is_uid)
                 parse_response(parsed_line.response);
 
                 if (mandatory_part_.empty())
-                    throw imap_error("Parsing failure.");
+                    throw imap_error("No mandatory part.", "Response=`" + parsed_line.response + "`.");
                 auto msg_no_token = mandatory_part_.front();
                 mandatory_part_.pop_front();
 
                 if (mandatory_part_.empty())
-                    throw imap_error("Parsing failure.");
+                    throw imap_error("No mandatory part.", "Response=`" + parsed_line.response + "`.");
                 auto fetch_token = mandatory_part_.front();
                 if (!iequals(fetch_token->atom, "FETCH"))
-                    throw imap_error("Parsing failure.");
+                    throw imap_error("Parsing failure.", "Tag=`" + fetch_token->atom + "`.");
                 mandatory_part_.pop_front();
 
                 // Check the list with flags.
 
                 if (mandatory_part_.empty())
-                    throw imap_error("Parsing failure.");
+                    throw imap_error("No mandatory part.", "Response=`" + parsed_line.response + "`.");
                 auto flags_token_list = mandatory_part_.front();
                 if (flags_token_list->token_type != response_token_t::token_type_t::LIST)
-                    throw imap_error("Parsing failure.");
+                    throw imap_error("Expecting the list.", "Line=`" + line + "`.");
 
                 std::shared_ptr<response_token_t> uid_token = nullptr;
                 auto uid_token_it = flags_token_list->parenthesized_list.begin();
@@ -670,7 +716,7 @@ void imap::remove(unsigned long message_no, bool is_uid)
                     {
                         uid_token_it++;
                         if (uid_token_it == flags_token_list->parenthesized_list.end())
-                            throw imap_error("Parsing failure.");
+                            throw imap_error("No UID.", "");
                         uid_token = *uid_token_it;
                         break;
                     }
@@ -681,19 +727,19 @@ void imap::remove(unsigned long message_no, bool is_uid)
                 if (is_uid)
                 {
                     if (uid_token == nullptr)
-                        throw imap_error("Parsing failure.");
+                        throw imap_error("No UID.", "");
                     msg_no_token = uid_token;
                 }
 
                 if (msg_no_token->token_type != response_token_t::token_type_t::ATOM || stoul(msg_no_token->atom) != message_no)
-                    throw imap_error("Deleting message failure.");
+                    throw imap_error("Deleting message failure.", "");
 
                 continue;
             }
             else if (parsed_line.tag == to_string(tag_))
             {
                 if (!parsed_line.result.has_value() || parsed_line.result.value() != tag_result_response_t::OK)
-                    throw imap_error("Deleting message failure.");
+                    throw imap_error("Deleting message failure.", "");
                 else
                 {
                     reset_response_parser();
@@ -702,23 +748,23 @@ void imap::remove(unsigned long message_no, bool is_uid)
                     tag_result_response_t parsed_line = parse_tag_result(line);
 
                     if (!iequals(parsed_line.tag, to_string(tag_)))
-                        throw imap_error("Parsing failure.");
+                        throw imap_error("Incorrect tag.", "Tag=`" + parsed_line.tag + "`.");
                     if (!parsed_line.result.has_value() || parsed_line.result.value() != tag_result_response_t::OK)
-                        throw imap_error("Deleting message failure.");
+                        throw imap_error("Deleting message failure.", "");
                 }
                 has_more = false;
             }
             else
-                throw imap_error("Parsing failure.");
+                throw imap_error("Incorrect tag.", "Tag=`" + parsed_line.tag + "`.");
         }
     }
-    catch (const invalid_argument&)
+    catch (const invalid_argument& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
-    catch (const out_of_range&)
+    catch (const out_of_range& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
 }
 
@@ -738,16 +784,16 @@ void imap::search(const list<imap::search_condition_t>& conditions, list<unsigne
 
 bool imap::create_folder(const string& folder_name)
 {
-    dlg_->send(format("CREATE " + QUOTED_STRING_SEPARATOR + folder_name + QUOTED_STRING_SEPARATOR));
+    dlg_->send(format("CREATE " + to_astring(folder_name)));
 
     string line = dlg_->receive();
     tag_result_response_t parsed_line = parse_tag_result(line);
     if (parsed_line.tag != to_string(tag_))
-        throw imap_error("Parsing failure.");
+        throw imap_error("Incorrect tag.", "Tag=`" + parsed_line.tag + "`.");
     if (parsed_line.result.value() == tag_result_response_t::NO)
         return false;
     if (parsed_line.result.value() != tag_result_response_t::OK)
-        throw imap_error("Creating folder failure.");
+        throw imap_error("Creating folder failure.", "Response=`" + parsed_line.response + "`.");
     return true;
 
 }
@@ -764,8 +810,7 @@ bool imap::create_folder(const list<string>& folder_name)
 auto imap::list_folders(const string& folder_name) -> mailbox_folder_t
 {
     string delim = folder_delimiter();
-    dlg_->send(format("LIST " + QUOTED_STRING_SEPARATOR + QUOTED_STRING_SEPARATOR + TOKEN_SEPARATOR_STR + QUOTED_STRING_SEPARATOR + folder_name + "*" +
-        QUOTED_STRING_SEPARATOR));
+    dlg_->send(format("LIST " + QUOTED_STRING_SEPARATOR + QUOTED_STRING_SEPARATOR + TOKEN_SEPARATOR_STR + to_astring(folder_name + "*")));
     mailbox_folder_t mailboxes;
 
     bool has_more = true;
@@ -781,10 +826,10 @@ auto imap::list_folders(const string& folder_name) -> mailbox_folder_t
                 auto token = mandatory_part_.front();
                 mandatory_part_.pop_front();
                 if (!iequals(token->atom, "LIST"))
-                    throw imap_error("Listing folders failure.");
+                    throw imap_error("Expecting the list atom.", "Atom=`" + token->atom + "`.");
 
                 if (mandatory_part_.size() < 3)
-                    throw imap_error("Parsing failure.");
+                    throw imap_error("Listing folders failure.", "");
                 auto found_folder = mandatory_part_.begin();
                 found_folder++; found_folder++;
                 if (found_folder != mandatory_part_.end() && (*found_folder)->token_type == response_token_t::token_type_t::ATOM)
@@ -802,7 +847,7 @@ auto imap::list_folders(const string& folder_name) -> mailbox_folder_t
                     }
                 }
                 else
-                    throw imap_error("Parsing failure.");
+                    throw imap_error("Parsing failure.", "Line=`" + line + "`.");
             }
             else if (parsed_line.tag == to_string(tag_))
             {
@@ -811,13 +856,13 @@ auto imap::list_folders(const string& folder_name) -> mailbox_folder_t
             reset_response_parser();
         }
     }
-    catch (const invalid_argument&)
+    catch (const invalid_argument& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
-    catch (const out_of_range&)
+    catch (const out_of_range& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
 
     return mailboxes;
@@ -834,16 +879,16 @@ auto imap::list_folders(const list<string>& folder_name) -> mailbox_folder_t
 
 bool imap::delete_folder(const string& folder_name)
 {
-    dlg_->send(format("DELETE " + QUOTED_STRING_SEPARATOR + folder_name + QUOTED_STRING_SEPARATOR));
+    dlg_->send(format("DELETE " + to_astring(folder_name)));
 
     string line = dlg_->receive();
     tag_result_response_t parsed_line = parse_tag_result(line);
     if (parsed_line.tag != to_string(tag_))
-        throw imap_error("Parsing failure.");
+        throw imap_error("Incorrect tag.", "Tag=`" + parsed_line.tag + "`.");
     if (parsed_line.result.value() == tag_result_response_t::NO)
         return false;
     if (parsed_line.result.value() != tag_result_response_t::OK)
-        throw imap_error("Deleting folder failure.");
+        throw imap_error("Deleting folder failure.", "Line=`" + line + "`.");
     return true;
 }
 
@@ -858,17 +903,16 @@ bool imap::delete_folder(const list<string>& folder_name)
 
 bool imap::rename_folder(const string& old_name, const string& new_name)
 {
-    dlg_->send(format("RENAME " + QUOTED_STRING_SEPARATOR + old_name + QUOTED_STRING_SEPARATOR + TOKEN_SEPARATOR_STR + QUOTED_STRING_SEPARATOR + new_name +
-        QUOTED_STRING_SEPARATOR));
+    dlg_->send(format("RENAME " + to_astring(old_name) + TOKEN_SEPARATOR_STR + to_astring(new_name)));
 
     string line = dlg_->receive();
     tag_result_response_t parsed_line = parse_tag_result(line);
     if (parsed_line.tag != to_string(tag_))
-        throw imap_error("Parsing failure.");
+        throw imap_error("Incorrect tag.", "Tag=`" + parsed_line.tag + "`.");
     if (parsed_line.result.value() == tag_result_response_t::NO)
         return false;
     if (parsed_line.result.value() != tag_result_response_t::OK)
-        throw imap_error("Renaming folder failure.");
+        throw imap_error("Renaming folder failure.", "Line=`" + line + "`.");
     return true;
 }
 
@@ -889,16 +933,18 @@ string imap::connect()
     tag_result_response_t parsed_line = parse_tag_result(line);
 
     if (parsed_line.tag != UNTAGGED_RESPONSE)
-        throw imap_error("Parsing failure.");
+        throw imap_error("Incorrect tag.", "Tag=`" + parsed_line.tag + "`.");
     if (!parsed_line.result.has_value() || parsed_line.result.value() != tag_result_response_t::OK)
-        throw imap_error("Connection to server failure.");
+        throw imap_error("Connection to server failure.", "Line=`" + line + "`.");
     return parsed_line.response;
 }
 
 
 void imap::auth_login(const string& username, const string& password)
 {
-    auto cmd = format("LOGIN " + username + TOKEN_SEPARATOR_STR + password);
+    auto user_esc = to_astring(username);
+    auto pass_esc = to_astring(password);
+    auto cmd = format("LOGIN " + user_esc + TOKEN_SEPARATOR_STR + pass_esc);
     dlg_->send(cmd);
 
     bool has_more = true;
@@ -910,9 +956,9 @@ void imap::auth_login(const string& username, const string& password)
         if (parsed_line.tag == UNTAGGED_RESPONSE)
             continue;
         if (parsed_line.tag != to_string(tag_))
-            throw imap_error("Parsing failure.");
+            throw imap_error("Incorrect tag.", "Tag=`" + parsed_line.tag + "`.");
         if (parsed_line.result.value() != tag_result_response_t::OK)
-            throw imap_error("Authentication failure.");
+            throw imap_error("Authentication failure.", "line=`" + line + "`.");
 
         has_more = false;
     }
@@ -940,8 +986,9 @@ void imap::search(const string& conditions, list<unsigned long>& results, bool w
                 parse_response(parsed_line.response);
 
                 auto search_token = mandatory_part_.front();
+                // ignore other responses, although not sure whether this is by the rfc or not
                 if (search_token->token_type == response_token_t::token_type_t::ATOM && !iequals(search_token->atom, "SEARCH"))
-                    throw imap_error("Search mailbox failure.");
+                    continue;
                 mandatory_part_.pop_front();
 
                 for (auto it = mandatory_part_.begin(); it != mandatory_part_.end(); it++)
@@ -949,30 +996,30 @@ void imap::search(const string& conditions, list<unsigned long>& results, bool w
                     {
                         const unsigned long idx = stoul((*it)->atom);
                         if (idx == 0)
-                            throw imap_error("Parsing failure.");
+                            throw imap_error("Incorrect message id.", "Line=`" + line + "`.");
                         results.push_back(idx);
                     }
             }
             else if (parsed_line.tag == to_string(tag_))
             {
                 if (parsed_line.result.value() != tag_result_response_t::OK)
-                    throw imap_error("Search mailbox failure.");
+                    throw imap_error("Search mailbox failure.", "Line=`" + line + "`.");
 
                 has_more = false;
             }
             else
             {
-                throw imap_error("Parsing failure.");
+                throw imap_error("Incorrect tag parsed.", "Tag=`" + parsed_line.tag + "`.");
             }
         }
     }
-    catch (const invalid_argument&)
+    catch (const invalid_argument& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
-    catch (const out_of_range&)
+    catch (const out_of_range& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
     reset_response_parser();
 }
@@ -994,21 +1041,21 @@ string imap::folder_delimiter()
                 {
                     parse_response(parsed_line.response);
                     if (!iequals(mandatory_part_.front()->atom, "LIST"))
-                        throw imap_error("Determining folder delimiter failure.");
+                        throw imap_error("Incorrect atom parsed.", "Line=`" + line + "`.");
                     mandatory_part_.pop_front();
 
                     if (mandatory_part_.size() < 3)
-                        throw imap_error("Determining folder delimiter failure.");
+                        throw imap_error("Determining folder delimiter failure.", "");
                     auto it = mandatory_part_.begin();
                     if ((*(++it))->token_type != response_token_t::token_type_t::ATOM)
-                        throw imap_error("Determining folder delimiter failure.");
+                        throw imap_error("Incorrect atom parsed.", "");
                     folder_delimiter_ = trim_copy_if((*it)->atom, [](char c ){ return c == QUOTED_STRING_SEPARATOR_CHAR; });
                     reset_response_parser();
                 }
                 else if (parsed_line.tag == to_string(tag_))
                 {
                     if (parsed_line.result.value() != tag_result_response_t::OK)
-                        throw imap_error("Determining folder delimiter failure.");
+                        throw imap_error("Determining folder delimiter failure.", "Line=`" + line + "`.");
 
                     has_more = false;
                 }
@@ -1016,13 +1063,13 @@ string imap::folder_delimiter()
         }
         return folder_delimiter_;
     }
-    catch (const invalid_argument&)
+    catch (const invalid_argument& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
-    catch (const out_of_range&)
+    catch (const out_of_range& exc)
     {
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", exc.what());
     }
 }
 
@@ -1031,7 +1078,7 @@ auto imap::parse_tag_result(const string& line) const -> tag_result_response_t
 {
     string::size_type tag_pos = line.find(TOKEN_SEPARATOR_STR);
     if (tag_pos == string::npos)
-        throw imap_error("Parsing failure.");
+        throw imap_error("Parsing failure.", "");
     string tag = line.substr(0, tag_pos);
 
     string::size_type result_pos = string::npos;
@@ -1076,7 +1123,7 @@ void imap::parse_response(const string& response)
     {
         token_list = optional_part_state_ ? find_last_token_list(optional_part_) : find_last_token_list(mandatory_part_);
         if (token_list->back()->token_type == response_token_t::token_type_t::LITERAL && literal_bytes_read_ > token_list->back()->literal.size())
-            throw imap_error("Parser failure.");
+            throw imap_error("Parser failure.", "");
         unsigned long literal_size = stoul(token_list->back()->literal_size);
         if (literal_bytes_read_ + response.size() < literal_size)
         {
@@ -1109,7 +1156,7 @@ void imap::parse_response(const string& response)
                 else
                 {
                     if (optional_part_state_)
-                        throw imap_error("Parser failure.");
+                        throw imap_error("Parser failure.", "");
 
                     optional_part_state_ = true;
                 }
@@ -1123,7 +1170,7 @@ void imap::parse_response(const string& response)
                 else
                 {
                     if (!optional_part_state_)
-                        throw imap_error("Parser failure.");
+                        throw imap_error("Parser failure.", "");
 
                     optional_part_state_ = false;
                     atom_state_ = atom_state_t::NONE;
@@ -1154,7 +1201,7 @@ void imap::parse_response(const string& response)
                 else
                 {
                     if (parenthesis_list_counter_ == 0)
-                        throw imap_error("Parser failure.");
+                        throw imap_error("Parser failure.", "");
 
                     parenthesis_list_counter_--;
                     atom_state_ = atom_state_t::NONE;
@@ -1169,7 +1216,7 @@ void imap::parse_response(const string& response)
                 else
                 {
                     if (literal_state_ == string_literal_state_t::SIZE)
-                        throw imap_error("Parser failure.");
+                        throw imap_error("Parser failure.", "");
 
                     cur_token = make_shared<response_token_t>();
                     cur_token->token_type = response_token_t::token_type_t::LITERAL;
@@ -1188,7 +1235,7 @@ void imap::parse_response(const string& response)
                 else
                 {
                     if (literal_state_ == string_literal_state_t::NONE)
-                        throw imap_error("Parser failure.");
+                        throw imap_error("Parser failure.", "");
 
                     literal_state_ = string_literal_state_t::WAITING;
                 }
@@ -1221,23 +1268,33 @@ void imap::parse_response(const string& response)
                     atom_state_ = atom_state_t::QUOTED;
                 }
                 else if (atom_state_ == atom_state_t::QUOTED)
-                    atom_state_ = atom_state_t::NONE;
+                {
+                    // The backslash and a double quote within an atom is the double quote only.
+                    if (token_list->back()->atom.back() != codec::BACKSLASH_CHAR)
+                        atom_state_ = atom_state_t::NONE;
+                    else
+                        token_list->back()->atom.back() = ch;
+                }
             }
             break;
 
             default:
             {
+                // Double backslash in an atom is translated to the single backslash.
+                if (ch == codec::BACKSLASH_CHAR && atom_state_ == atom_state_t::QUOTED && token_list->back()->atom.back() == codec::BACKSLASH_CHAR)
+                    break;
+
                 if (literal_state_ == string_literal_state_t::SIZE)
                 {
                     if (!isdigit(ch))
-                        throw imap_error("Parser failure.");
+                        throw imap_error("Parser failure.", "");
 
                     cur_token->literal_size += ch;
                 }
                 else if (literal_state_ == string_literal_state_t::WAITING)
                 {
                     // no characters allowed after the right brace, crlf is required
-                    throw imap_error("Parser failure.");
+                    throw imap_error("Parser failure.", "");
                 }
                 else
                 {
@@ -1254,6 +1311,7 @@ void imap::parse_response(const string& response)
             }
         }
     }
+
     if (literal_state_ == string_literal_state_t::WAITING)
         literal_state_ = string_literal_state_t::READING;
 }
@@ -1332,7 +1390,7 @@ imaps::imaps(const string& hostname, unsigned port, milliseconds timeout) : imap
     ssl_options_ =
         {
             boost::asio::ssl::context::sslv23,
-            boost::asio::ssl::verify_peer
+            boost::asio::ssl::verify_none
         };
 }
 
@@ -1368,9 +1426,9 @@ void imaps::start_tls()
     string line = dlg_->receive();
     tag_result_response_t parsed_line = parse_tag_result(line);
     if (parsed_line.tag == UNTAGGED_RESPONSE)
-        throw imap_error("Bad server response.");
+        throw imap_error("Bad server response.", "");
     if (parsed_line.result.value() != tag_result_response_t::OK)
-        throw imap_error("Start TLS refused by server.");
+        throw imap_error("Start TLS refused by server.", "");
 
     switch_to_ssl();
 }
@@ -1382,4 +1440,14 @@ void imaps::switch_to_ssl()
 }
 
 
+imap_error::imap_error(const string& msg, const string& details) : dialog_error(msg, details)
+{
+}
+
+
+imap_error::imap_error(const char* msg, const string& details) : dialog_error(msg, details)
+{
+}
+
+
 } // namespace mailio
