@@ -27,16 +27,6 @@ mail_address::mail_address(const string_t& mail_name, const string& mail_address
 }
 
 
-mail_address::mail_address(const string& mail_name, const string& mail_address)
-{
-    if (codec::is_utf8_string(mail_name))
-        name = string_t(mail_name, codec::CHARSET_UTF8);
-    else
-        name = string_t(mail_name, codec::CHARSET_ASCII);
-    address = mail_address;
-}
-
-
 bool mail_address::empty() const
 {
     return name.buffer.empty() && address.empty();
