@@ -12,6 +12,7 @@ copy at http://www.freebsd.org/copyright/freebsd-license.html.
 
 
 #include <string>
+#include <string_view>
 #include <fstream>
 #include <sstream>
 #include <map>
@@ -25,13 +26,16 @@ copy at http://www.freebsd.org/copyright/freebsd-license.html.
 #include <mailio/bit8.hpp>
 #include <mailio/binary.hpp>
 #include <mailio/q_codec.hpp>
+#include <mailio/percent.hpp>
 #include <mailio/mime.hpp>
 
 
 using std::string;
+using std::string_view;
 #if defined(__cpp_char8_t)
 using std::u8string;
 #endif
+using std::to_string;
 using std::ifstream;
 using std::stringstream;
 using std::pair;
@@ -39,20 +43,30 @@ using std::multimap;
 using std::vector;
 using std::map;
 using std::get;
+using boost::trim_left_if;
+using boost::trim_right_if;
 using boost::regex;
 using boost::regex_match;
 using boost::smatch;
+using boost::match_flag_type;
+using boost::match_results;
 using boost::to_lower_copy;
 using boost::trim_copy;
 using boost::trim;
 using boost::trim_right;
 using boost::iequals;
+using boost::algorithm::is_any_of;
 
 
 namespace mailio
 {
 
-
+const string mime::CONTENT_ID_HEADER{"Content-ID"};
+const string mime::ADDRESS_BEGIN_STR(1, ADDRESS_BEGIN_CHAR);
+const string mime::ADDRESS_END_STR(1, ADDRESS_END_CHAR);
+const string mime::MESSAGE_ID_REGEX = "([a-zA-Z0-9\\!#\\$%&'\\*\\+\\-\\./=\\?\\^\\_`\\{\\|\\}\\~]+)"
+    "\\@([a-zA-Z0-9\\!#\\$%&'\\*\\+\\-\\./=\\?\\^\\_`\\{\\|\\}\\~]+)";
+const string mime::MESSAGE_ID_REGEX_NS = "([a-zA-Z0-9\\!#\\$%&'\\*\\+\\-\\./=\\?\\^\\_`\\{\\|\\}\\~\\@\\\\ \\t<\\>]*)";
 const string mime::content_type_t::ATTR_CHARSET{"charset"};
 const string mime::content_type_t::ATTR_BOUNDARY{"boundary"};
 
@@ -96,6 +110,7 @@ const string mime::NEW_LINE_INDENT{"  "};
 const string mime::HEADER_SEPARATOR_STR{": "};
 const string mime::NAME_VALUE_SEPARATOR_STR{"="};
 const string mime::ATTRIBUTES_SEPARATOR_STR{"; "};
+const string mime::ATTRIBUTE_CONTINUATION_INDICATOR{"*"};
 const string mime::ATTRIBUTE_NAME{"name"};
 const string mime::ATTRIBUTE_FILENAME{"filename"};
 const string mime::BOUNDARY_DELIMITER(2, '-');
@@ -103,14 +118,14 @@ const string mime::QTEXT{"\t !#$%&'()*+,-.@/:;<=>?[]^_`{|}~"};
 // exluded double quote and backslash characters from the header name
 const regex mime::HEADER_NAME_REGEX{R"(([a-zA-Z0-9\!#\$%&'\(\)\*\+\-\./;\<=\>\?@\[\\\]\^\_`\{\|\}\~]+))"};
 const regex mime::HEADER_VALUE_REGEX{R"(([a-zA-Z0-9\ \t\!\"#\$%&'\(\)\*\+\,\-\./:;\<=\>\?@\[\\\]\^\_`\{\|\}\~]+))"};
-const string mime::CONTENT_ATTR_ALPHABET{"!#$%&*+-.^_`|~"};
+const string mime::CONTENT_ATTR_ALPHABET{"!#$%&'*+-.^_`|~"};
 const string mime::CONTENT_HEADER_VALUE_ALPHABET{"!#$%&*+-./^_`|~"};
 
 
-mime::mime() : version_("1.0"), line_policy_(codec::line_len_policy_t::RECOMMENDED),
-    decoder_line_policy_(codec::line_len_policy_t::RECOMMENDED), strict_mode_(false), strict_codec_mode_(false),
-    header_codec_(header_codec_t::UTF8), content_type_(media_type_t::NONE, ""), encoding_(content_transfer_encoding_t::NONE),
-    disposition_(content_disposition_t::NONE), parsing_header_(true), mime_status_(mime_parsing_status_t::NONE)
+mime::mime() : version_("1.0"), line_policy_(codec::line_len_policy_t::MANDATORY),
+    strict_mode_(false), strict_codec_mode_(false), header_codec_(header_codec_t::UTF8), content_type_(media_type_t::NONE, ""),
+    encoding_(content_transfer_encoding_t::NONE), disposition_(content_disposition_t::NONE), parsing_header_(true),
+    mime_status_(mime_parsing_status_t::NONE)
 {
 }
 
@@ -118,7 +133,7 @@ mime::mime() : version_("1.0"), line_policy_(codec::line_len_policy_t::RECOMMEND
 void mime::format(string& mime_str, bool dot_escape) const
 {
     if (!boundary_.empty() && content_type_.type != media_type_t::MULTIPART)
-        throw mime_error("Formatting failure, non multipart message with boundary.");
+        throw mime_error("Formatting failure, non multipart message with boundary.", "");
 
     mime_str += format_header() + codec::END_OF_LINE;
     string content = format_content(dot_escape);
@@ -128,7 +143,9 @@ void mime::format(string& mime_str, bool dot_escape) const
     {
         if (!content.empty())
             mime_str += codec::END_OF_LINE;
-        // recursively format mime parts
+
+        // Recursively format mime parts.
+
         for (auto& p : parts_)
         {
             string p_str;
@@ -182,8 +199,8 @@ void mime::parse(const u8string& mime_string, bool dot_escape)
 
 mime& mime::parse_by_line(const string& line, bool dot_escape)
 {
-    if (line.length() > string::size_type(decoder_line_policy_))
-        throw mime_error("Line policy overflow in a header.");
+    if (line.length() > string::size_type(line_policy_))
+        throw mime_error("Line policy overflow in a header.", "Line is `" + line + "`.");
 
     // mark end of header and parse it
     if (parsing_header_ && line.empty())
@@ -214,7 +231,7 @@ mime& mime::parse_by_line(const string& line, bool dot_escape)
                     if (!parts_.empty())
                         parts_.back().parse_by_line(codec::END_OF_LINE);
                     mime m;
-                    m.line_policy(line_policy_, decoder_line_policy_);
+                    m.line_policy(line_policy_);
                     m.strict_codec_mode(strict_codec_mode_);
                     parts_.push_back(m);
                 }
@@ -250,7 +267,7 @@ mime& mime::parse_by_line(const string& line, bool dot_escape)
 void mime::content_type(const content_type_t& cont_type)
 {
     if (cont_type.type != media_type_t::NONE && cont_type.subtype.empty())
-        throw mime_error("Bad content type.");
+        throw mime_error("Bad content type.", "Media type is none, subtype is `" + cont_type.subtype + "`.");
     content_type_ = cont_type;
 }
 
@@ -269,13 +286,31 @@ mime::content_type_t mime::content_type() const
 }
 
 
-void mime::name(const string& mime_name)
+void mime::content_id(string id)
+{
+    const regex r(strict_mode_ ? MESSAGE_ID_REGEX : MESSAGE_ID_REGEX_NS);
+    smatch m;
+
+    if (regex_match(id, m, r))
+        content_id_ = id;
+    else
+        throw mime_error("Invalid content ID.", "Content ID is `" + id + "`.");
+}
+
+
+string mime::content_id() const
+{
+    return content_id_;
+}
+
+
+void mime::name(const string_t& mime_name)
 {
     name_ = mime_name;
 }
 
 
-string mime::name() const
+string_t mime::name() const
 {
     return name_;
 }
@@ -349,10 +384,15 @@ vector<mime> mime::parts() const
 }
 
 
-void mime::line_policy(codec::line_len_policy_t encoder_line_policy, codec::line_len_policy_t decoder_line_policy)
+void mime::line_policy(codec::line_len_policy_t encoder_line_policy, codec::line_len_policy_t /*decoder_line_policy*/)
 {
     line_policy_ = encoder_line_policy;
-    decoder_line_policy_ = decoder_line_policy;
+}
+
+
+void mime::line_policy(codec::line_len_policy_t line_policy)
+{
+    line_policy_ = line_policy;
 }
 
 
@@ -362,12 +402,6 @@ codec::line_len_policy_t mime::line_policy() const
 }
 
 
-codec::line_len_policy_t mime::decoder_line_policy() const
-{
-    return decoder_line_policy_;
-}
-
-
 void mime::strict_mode(bool mode)
 {
     strict_mode_ = mode;
@@ -404,9 +438,64 @@ mime::header_codec_t mime::header_codec() const
 }
 
 
+string mime::format_many_ids(const vector<string>& ids, const string& header_name) const
+{
+    string ids_str;
+    for (auto s = ids.begin(); s != ids.end(); s++)
+    {
+        ids_str += ADDRESS_BEGIN_STR + *s + ADDRESS_END_STR;
+        if (s != ids.end() - 1)
+            ids_str += codec::SPACE_STR;
+    }
+
+    string::size_type l1p = static_cast<string::size_type>(line_policy_) - header_name.length() - HEADER_SEPARATOR_STR.length();
+    bit7 b7(l1p, static_cast<string::size_type>(line_policy_));
+    vector<string> ids_enc = b7.encode(ids_str);
+    string ids_folded = ids_enc.at(0) + codec::END_OF_LINE;
+    ids_folded += fold_header_line(ids_enc);
+    return ids_folded;
+}
+
+
+string mime::format_many_ids(const string& id, const string& header_name) const
+{
+    return format_many_ids(vector<string>{id}, header_name);
+}
+
+
+vector<string> mime::parse_many_ids(const string& ids) const
+{
+    if (!strict_mode_)
+        return vector<string>{ids};
+
+    vector<string> idv;
+    auto start = ids.cbegin();
+    auto end = ids.cend();
+    match_flag_type flags = boost::match_default | boost::match_not_null;
+    match_results<string::const_iterator> tokens;
+    bool all_tokens_parsed = false;
+    const string ANGLED_MESSAGE_ID_REGEX = codec::LESS_THAN_STR + MESSAGE_ID_REGEX + codec::GREATER_THAN_STR;
+    const regex rgx(ANGLED_MESSAGE_ID_REGEX);
+    while (regex_search(start, end, tokens, rgx, flags))
+    {
+        string id = tokens[0];
+        trim_left_if(id, is_any_of(codec::LESS_THAN_STR));
+        trim_right_if(id, is_any_of(codec::GREATER_THAN_STR));
+        idv.push_back(id);
+        start = tokens[0].second;
+        all_tokens_parsed = (start == end);
+    }
+
+    if (!all_tokens_parsed)
+        throw mime_error("Parsing failure of the ID.", "ID is `" + ids + "`.");
+
+    return idv;
+}
+
+
 string mime::format_header() const
 {
-    return format_content_type() + format_transfer_encoding() + format_content_disposition();
+    return format_content_type() + format_transfer_encoding() + format_content_disposition() + format_content_id();
 }
 
 
@@ -417,7 +506,7 @@ string mime::format_content(bool dot_escape) const
     {
         case content_transfer_encoding_t::BASE_64:
         {
-            base64 b(line_policy_, decoder_line_policy_);
+            base64 b(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
             b.strict_mode(strict_codec_mode_);
             content_lines = b.encode(content_);
             break;
@@ -425,7 +514,7 @@ string mime::format_content(bool dot_escape) const
 
         case content_transfer_encoding_t::QUOTED_PRINTABLE:
         {
-            quoted_printable qp(line_policy_, decoder_line_policy_);
+            quoted_printable qp(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
             qp.strict_mode(strict_codec_mode_);
             content_lines = qp.encode(content_);
             break;
@@ -434,7 +523,7 @@ string mime::format_content(bool dot_escape) const
         // TODO: check how to handle 8bit chars, see [rfc 2045, section 2.8]
         case content_transfer_encoding_t::BIT_8:
         {
-            bit8 b8(line_policy_, decoder_line_policy_);
+            bit8 b8(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
             b8.strict_mode(strict_codec_mode_);
             content_lines = b8.encode(content_);
             break;
@@ -444,7 +533,7 @@ string mime::format_content(bool dot_escape) const
         case content_transfer_encoding_t::BIT_7:
         case content_transfer_encoding_t::NONE:
         {
-            bit7 b7(line_policy_, decoder_line_policy_);
+            bit7 b7(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
             b7.strict_mode(strict_codec_mode_);
             content_lines = b7.encode(content_);
             break;
@@ -453,7 +542,7 @@ string mime::format_content(bool dot_escape) const
         case content_transfer_encoding_t::BINARY:
         {
             // TODO: probably bug when `\0` is part of the content
-            binary b(line_policy_, decoder_line_policy_);
+            binary b(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
             b.strict_mode(strict_codec_mode_);
             content_lines = b.encode(content_);
             break;
@@ -482,15 +571,10 @@ string mime::format_content_type() const
         line += CONTENT_TYPE_HEADER + HEADER_SEPARATOR_STR + mime_type_as_str(content_type_.type) + CONTENT_SUBTYPE_SEPARATOR + content_type_.subtype;
         if (!content_type_.charset.empty())
             line += ATTRIBUTES_SEPARATOR_STR + content_type_t::ATTR_CHARSET + NAME_VALUE_SEPARATOR_STR + content_type_.charset;
-        if (!name_.empty())
+        if (!name_.buffer.empty())
         {
-            string mn = format_mime_name(name_);
-            const string::size_type line_new_size = line.size() + ATTRIBUTES_SEPARATOR_STR.size() + ATTRIBUTE_NAME.size() + NAME_VALUE_SEPARATOR_STR.size() +
-                sizeof(codec::QUOTE_CHAR) + mn.size() + sizeof(codec::QUOTE_CHAR);
-            line += ATTRIBUTES_SEPARATOR_STR;
-            if (line_new_size >= string::size_type(line_policy_) - codec::END_OF_LINE.length())
-                line += codec::END_OF_LINE + NEW_LINE_INDENT;
-            line += ATTRIBUTE_NAME + NAME_VALUE_SEPARATOR_STR + codec::QUOTE_STR + mn + codec::QUOTE_STR;
+            string attrs = split_attributes(ATTRIBUTE_NAME, name_);
+            line += ATTRIBUTES_SEPARATOR_STR + codec::END_OF_LINE + attrs;
         }
         if (!boundary_.empty())
             line += ATTRIBUTES_SEPARATOR_STR + content_type_t::ATTR_BOUNDARY + NAME_VALUE_SEPARATOR_STR + codec::QUOTE_CHAR + boundary_ + codec::QUOTE_CHAR;
@@ -535,56 +619,32 @@ string mime::format_transfer_encoding() const
 
 string mime::format_content_disposition() const
 {
-    string line;
+    string disp;
+    if (disposition_ == content_disposition_t::NONE)
+        return "";
+    else if (disposition_ == content_disposition_t::ATTACHMENT)
+        disp = CONTENT_DISPOSITION_ATTACHMENT;
+    else if (disposition_ == content_disposition_t::INLINE)
+        disp = CONTENT_DISPOSITION_INLINE;
 
-    switch (disposition_)
-    {
-        case content_disposition_t::ATTACHMENT:
-        {
-            string name = format_mime_name(name_);
-            line += CONTENT_DISPOSITION_HEADER + HEADER_SEPARATOR_STR + CONTENT_DISPOSITION_ATTACHMENT + ATTRIBUTES_SEPARATOR_STR;
-            const string::size_type line_new_size = CONTENT_DISPOSITION_HEADER.size() + HEADER_SEPARATOR_STR.size() + CONTENT_DISPOSITION_ATTACHMENT.size() +
-                ATTRIBUTE_FILENAME.size() + sizeof(codec::EQUAL_CHAR) + sizeof(codec::QUOTE_CHAR) + name.size() + sizeof(codec::QUOTE_CHAR) +
-                codec::END_OF_LINE.size();
-            if (line_new_size >= string::size_type(line_policy_) - codec::END_OF_LINE.length())
-                line += codec::END_OF_LINE + NEW_LINE_INDENT;
-            line += ATTRIBUTE_FILENAME + NAME_VALUE_SEPARATOR_STR + codec::QUOTE_CHAR + name + codec::QUOTE_STR + codec::END_OF_LINE;
-
-            break;
-        }
-
-        case content_disposition_t::INLINE:
-        {
-            string name = format_mime_name(name_);
-            line += CONTENT_DISPOSITION_HEADER + HEADER_SEPARATOR_STR + CONTENT_DISPOSITION_INLINE + ATTRIBUTES_SEPARATOR_STR;
-            const string::size_type line_new_size = CONTENT_DISPOSITION_HEADER.size() + HEADER_SEPARATOR_STR.size() + CONTENT_DISPOSITION_INLINE.size() +
-                ATTRIBUTE_FILENAME.size() + NAME_VALUE_SEPARATOR_STR.size() + sizeof(codec::QUOTE_CHAR) + name.size() + sizeof(codec::QUOTE_CHAR) + codec::END_OF_LINE.size();
-            if (line_new_size >= string::size_type(line_policy_) - codec::END_OF_LINE.length())
-                line += codec::END_OF_LINE + NEW_LINE_INDENT;
-            line += ATTRIBUTE_FILENAME + NAME_VALUE_SEPARATOR_STR + codec::QUOTE_CHAR + name + codec::QUOTE_CHAR + codec::END_OF_LINE;
-
-            break;
-        }
-
-        default:
-            break;
-    }
-
-    return line;
+    string attrs = split_attributes(ATTRIBUTE_FILENAME, name_);
+    return CONTENT_DISPOSITION_HEADER + HEADER_SEPARATOR_STR + disp + ATTRIBUTES_SEPARATOR_STR + codec::END_OF_LINE + attrs + codec::END_OF_LINE;
 }
 
 
-string mime::format_mime_name(const string& name) const
+string mime::format_content_id() const
 {
-    if (codec::is_utf8_string(name))
-    {
-        // if the attachment name exceeds mandatory length, the rest is discarded
-        q_codec qc(codec::line_len_policy_t::MANDATORY, decoder_line_policy_);
-        vector<string> hdr = qc.encode(name, codec::CHARSET_UTF8, header_codec_);
-        return hdr.at(0);
-    }
+    return content_id_.empty() ? "" : CONTENT_ID_HEADER + HEADER_SEPARATOR_STR + format_many_ids(content_id_, CONTENT_ID_HEADER);
+}
 
-    return name;
+
+string mime::fold_header_line(const vector<string>& headers) const
+{
+    string hdr_str;
+    if (headers.size() > 1)
+        for (auto h = headers.begin() + 1; h != headers.end(); h++)
+            hdr_str += codec::SPACE_STR + codec::SPACE_STR + *h + codec::END_OF_LINE;
+    return hdr_str;
 }
 
 
@@ -618,7 +678,7 @@ void mime::parse_content()
     {
         case content_transfer_encoding_t::BASE_64:
         {
-            base64 b64(line_policy_, decoder_line_policy_);
+            base64 b64(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
             b64.strict_mode(strict_codec_mode_);
             content_ = b64.decode(parsed_body_);
             break;
@@ -626,7 +686,7 @@ void mime::parse_content()
 
         case content_transfer_encoding_t::QUOTED_PRINTABLE:
         {
-            quoted_printable qp(line_policy_, decoder_line_policy_);
+            quoted_printable qp(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
             qp.strict_mode(strict_codec_mode_);
             content_ = qp.decode(parsed_body_);
             break;
@@ -634,7 +694,7 @@ void mime::parse_content()
 
         case content_transfer_encoding_t::BIT_8:
         {
-            bit8 b8(line_policy_, decoder_line_policy_);
+            bit8 b8(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
             b8.strict_mode(strict_codec_mode_);
             content_ = b8.decode(parsed_body_);
             break;
@@ -643,7 +703,7 @@ void mime::parse_content()
         case content_transfer_encoding_t::BIT_7:
         case content_transfer_encoding_t::NONE:
         {
-            bit7 b7(line_policy_, decoder_line_policy_);
+            bit7 b7(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
             b7.strict_mode(strict_codec_mode_);
             content_ = b7.decode(parsed_body_);
             break;
@@ -651,13 +711,13 @@ void mime::parse_content()
 
         case content_transfer_encoding_t::BINARY:
         {
-            binary b(line_policy_, decoder_line_policy_);
+            binary b(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
             b.strict_mode(strict_codec_mode_);
             content_ = b.decode(parsed_body_);
             break;
         }
 
-        // default encoding is seven bit, so no `default` clause
+        // Default encoding is seven bit, so no `default` clause.
     }
 }
 
@@ -682,22 +742,20 @@ void mime::parse_header_line(const string& header_line)
         string media_subtype;
         attributes_t attributes;
         parse_content_type(header_value, media_type, media_subtype, attributes);
+        merge_attributes(attributes);
 
         content_type_.type = media_type;
         content_type_.subtype = to_lower_copy(media_subtype);
         attributes_t::iterator bound_it = attributes.find(content_type_t::ATTR_BOUNDARY);
         if (bound_it != attributes.end())
-            boundary_ = bound_it->second;
+            boundary_ = bound_it->second.buffer;
         attributes_t::iterator charset_it = attributes.find(content_type_t::ATTR_CHARSET);
         if (charset_it != attributes.end())
-            content_type_.charset = to_lower_copy(charset_it->second);
+            content_type_.charset = to_lower_copy(charset_it->second.buffer);
         attributes_t::iterator name_it = attributes.find(ATTRIBUTE_NAME);
         // name is set if not already set by content disposition
-        if (name_it != attributes.end() && name_.empty())
-        {
-            q_codec qc(line_policy_, decoder_line_policy_);
-            name_ = get<0>(qc.check_decode(name_it->second));
-        }
+        if (name_it != attributes.end() && name_.buffer.empty())
+            name_ = name_it->second;
     }
     else if (iequals(header_name, CONTENT_TRANSFER_ENCODING_HEADER))
     {
@@ -709,13 +767,17 @@ void mime::parse_header_line(const string& header_line)
         attributes_t attributes;
         parse_content_disposition(header_value, disposition_, attributes);
         merge_attributes(attributes);
+
         // filename is stored no matter if name is already set by content type
         attributes_t::iterator filename_it = attributes.find(ATTRIBUTE_FILENAME);
         if (filename_it != attributes.end())
-        {
-            q_codec qc(line_policy_, decoder_line_policy_);
-            name_ = get<0>(qc.check_decode(filename_it->second));
-        }
+            name_ = filename_it->second;
+    }
+    else if (iequals(header_name, CONTENT_ID_HEADER))
+    {
+        auto ids = parse_many_ids(header_value);
+        if (!ids.empty())
+            content_id_ = ids[0];
     }
 }
 
@@ -724,7 +786,7 @@ void mime::parse_header_name_value(const string& header_line, string& header_nam
 {
     string::size_type colon_pos = header_line.find(HEADER_SEPARATOR_CHAR);
     if (colon_pos == string::npos)
-        throw mime_error("Parsing failure of header line.");
+        throw mime_error("Parsing failure of header line.", "Header line is `" + header_line + "`.");
 
     header_name = header_line.substr(0, colon_pos);
     trim(header_name);
@@ -732,25 +794,21 @@ void mime::parse_header_name_value(const string& header_line, string& header_nam
     trim(header_value);
 
     if (header_name.empty())
-        throw mime_error("Parsing failure, header name or value empty: " + header_line);
+        throw mime_error("Parsing failure, header name empty.", "Header line parsed is `" + header_line + "`.");
     smatch m;
     if (!regex_match(header_name, m, HEADER_NAME_REGEX))
-        throw mime_error("Format failure of the header name `" + header_name + "`.");
+        throw mime_error("Format failure of the header name.", "Header line parsed is `" + header_line + "`.");
 
     if (header_value.empty())
     {
         if (strict_mode_)
-        {
-            throw mime_error("Parsing failure, header name or value empty: " + header_line);
-        }
+            throw mime_error("Parsing failure, header value empty.", "Header line parsed is `" + header_line + "`.");
         else
-        {
             return;
-        }
     }
 
     if (!codec::is_utf8_string(header_value) && !regex_match(header_value, m, HEADER_VALUE_REGEX))
-        throw mime_error("Format failure of the header value `" + header_value + "`.");
+        throw mime_error("Format failure of the header value", "Header value is `" + header_value + "`.");
 }
 
 
@@ -770,7 +828,7 @@ void mime::parse_content_type(const string& content_type_hdr, media_type_t& medi
         }
         else if (!isalpha(ch) && !isdigit(ch) && CONTENT_ATTR_ALPHABET.find(ch) == string::npos)
         {
-            throw mime_error("Parsing content type value failure.");
+            throw mime_error("Parsing content type value failure.", "Header content type header is `" + content_type_hdr + "`.");
         }
         else if (is_media_type)
         {
@@ -800,7 +858,7 @@ void mime::parse_content_transfer_encoding(const string& transfer_encoding_hdr,
     else if (iequals(header_value, CONTENT_TRANSFER_ENCODING_BINARY))
         encoding = content_transfer_encoding_t::BINARY;
     else if (strict_mode_)
-        throw mime_error("Parsing content transfer encoding failure.");
+        throw mime_error("Parsing content transfer encoding failure.", "Content transfer header is `" + transfer_encoding_hdr  + "`.");
     else
         encoding = content_transfer_encoding_t::BIT_7;
 }
@@ -817,7 +875,7 @@ void mime::parse_content_disposition(const string& content_disp_hdr, content_dis
     else
     {
         if (strict_mode_)
-            throw mime_error("Parsing content disposition failure.");
+            throw mime_error("Parsing content disposition failure.", "Content disposition header is `" + content_disp_hdr  + "`.");
         else
             disposition = content_disposition_t::ATTACHMENT;
     }
@@ -868,7 +926,8 @@ void mime::parse_header_value_attributes(const string& header, string& header_va
     state_t state = state_t::BEGIN;
     string attribute_name, attribute_value;
 
-    for (auto ch = header.begin(); ch != header.end(); ch++)
+    size_t char_pos = 0;
+    for (auto ch = header.begin(); ch != header.end(); ch++, char_pos++)
     {
         switch (state)
         {
@@ -881,7 +940,8 @@ void mime::parse_header_value_attributes(const string& header, string& header_va
                     header_value += *ch;
                 }
                 else
-                    throw mime_error("Parsing header value failure at `" + string(1, *ch) + "`.");
+                    throw mime_error("Parsing header value error.", "Syntax error at character `" + string(1, *ch) + "`, at position " + to_string(char_pos)
+                        + ".\nHeader is `" + header + "`.");
                 break;
 
             case state_t::VALUE:
@@ -890,7 +950,8 @@ void mime::parse_header_value_attributes(const string& header, string& header_va
                 else if (*ch == ATTRIBUTES_SEPARATOR_CHAR)
                     state = state_t::ATTR_BEGIN;
                 else
-                    throw mime_error("Parsing header value failure at `" + string(1, *ch) + "`.");
+                    throw mime_error("Parsing header value error.", "Syntax error at character `" + string(1, *ch) + "`, at position " + to_string(char_pos)
+                        + ".\nHeader is `" + header + "`.");
                 break;
 
             case state_t::ATTR_BEGIN:
@@ -904,7 +965,8 @@ void mime::parse_header_value_attributes(const string& header, string& header_va
                 else if (*ch == NAME_VALUE_SEPARATOR_CHAR)
                     state = state_t::ATTR_SEP;
                 else
-                    throw mime_error("Parsing attribute name failure at `" + string(1, *ch) + "`.");
+                    throw mime_error("Parsing attribute name error.", "Syntax error at character `" + string(1, *ch) + "`, at position " + to_string(char_pos)
+                        + ".\nHeader is `" + header + "`.");
                 break;
 
             case state_t::ATTR_NAME:
@@ -934,7 +996,8 @@ void mime::parse_header_value_attributes(const string& header, string& header_va
                     attribute_value += *ch;
                 }
                 else
-                    throw mime_error("Parsing attribute value failure at `" + string(1, *ch) + "`.");
+                    throw mime_error("Parsing attribute value error.", "Syntax error at character `" + string(1, *ch) + "`, at position " +
+                        to_string(char_pos) + ".\nHeader is `" + header + "`.");
                 break;
 
             case state_t::POST_EQUAL:
@@ -948,7 +1011,8 @@ void mime::parse_header_value_attributes(const string& header, string& header_va
                     attribute_value += *ch;
                 }
                 else
-                    throw mime_error("Parsing attribute value failure at `" + string(1, *ch) + "`.");
+                    throw mime_error("Parsing attribute value error.", "Syntax error at character `" + string(1, *ch) + "` at position " +
+                        to_string(char_pos) + ".\nHeader is `" + header + "`.");
                 break;
 
             case state_t::QATTR_VALUE_BEGIN:
@@ -959,9 +1023,12 @@ void mime::parse_header_value_attributes(const string& header, string& header_va
                 else if (*ch == codec::QUOTE_CHAR)
                     state = state_t::ATTR_VALUE_END;
                 else
-                    throw mime_error("Parsing attribute value failure at `" + string(1, *ch) + "`.");
+                    throw mime_error("Parsing attribute value error.", "Syntax error at character `" + string(1, *ch) + "` at position " +
+                        to_string(char_pos) + ".\nHeader is `" + header + "`.");
                 if (ch == header.end() - 1)
+                {
                     attributes[attribute_name] = attribute_value;
+                }
                 break;
 
             case state_t::ATTR_VALUE_BEGIN:
@@ -979,12 +1046,16 @@ void mime::parse_header_value_attributes(const string& header, string& header_va
                     attribute_value.clear();
                 }
                 else
-                    throw mime_error("Parsing attribute value failure at `" + string(1, *ch) + "`.");
+                    throw mime_error("Parsing attribute value error.", "Syntax error at character `" + string(1, *ch) + "` at position " + to_string(char_pos)
+                        + ".\nHeader is `" + header + "`.");
                 if (ch == header.end() - 1)
+                {
                     attributes[attribute_name] = attribute_value;
+                }
                 break;
 
             case state_t::ATTR_VALUE_END:
+            {
                 attributes[attribute_name] = attribute_value;
                 attribute_name.clear();
                 attribute_value.clear();
@@ -994,8 +1065,10 @@ void mime::parse_header_value_attributes(const string& header, string& header_va
                 else if (*ch == ATTRIBUTES_SEPARATOR_CHAR)
                     state = state_t::ATTR_BEGIN;
                 else
-                    throw mime_error("Parsing attribute value failure at `" + string(1, *ch) + "`.");
+                    throw mime_error("Parsing attribute value error.", "Syntax error at character `" + string(1, *ch) + "` at position " + to_string(char_pos)
+                        + ".\nHeader is `" + header + "`.");
                 break;
+            }
 
             case state_t::END:
                 return;
@@ -1004,57 +1077,161 @@ void mime::parse_header_value_attributes(const string& header, string& header_va
 }
 
 
+string mime::split_attributes(const string& attr_name, const string_t& attr_value) const
+{
+    // For the continuation like `filename*0=UTF-8'en-us'C%3A\\Program%20Files\\; ` there are five characters for the charset and two single quotes, because the
+    // language is not supported. There are in addition the equal. In total, this is nine characters.
+    const string::size_type CHARSET_LANG_EQ_SC = 9;
+    // Attribute name can be at most 10 parts enumerated from 0 to 9, with the continuation mark in between, so it adds two more characters to the line length.
+    const string::size_type PART_MARK = 2;
+    const string::size_type line1_policy = static_cast<string::size_type>(line_policy_) - (attr_name.length() + PART_MARK) - CHARSET_LANG_EQ_SC -
+        ATTRIBUTES_SEPARATOR_STR.length();
+    // Deduce two tab characters for the folding and two for double quotes.
+    const string::size_type FOLD_QUOTES = 4;
+    const string::size_type line_policy = static_cast<string::size_type>(line_policy_) - (attr_name.length() + PART_MARK) - FOLD_QUOTES -
+        ATTRIBUTES_SEPARATOR_STR.length();
+    vector<string> attr_parts;
+
+    if (attr_value.codec_type == codec::codec_t::ASCII)
+    {
+        bit7 b7(line1_policy, line_policy);
+        attr_parts = b7.encode(attr_value);
+    }
+    else if (attr_value.codec_type == codec::codec_t::QUOTED_PRINTABLE || attr_value.codec_type == codec::codec_t::BASE64)
+    {
+        q_codec qc(line1_policy, line_policy);
+        attr_parts = qc.encode(attr_value, attr_value.charset, attr_value.codec_type);
+    }
+    else if (attr_value.codec_type == codec::codec_t::PERCENT)
+    {
+        percent pc(line1_policy, line_policy);
+        attr_parts = pc.encode(attr_value, attr_value.charset);
+    }
+    else
+    {
+        throw codec_error("Attribute codec not supported.");
+    }
+
+    // Only one part means there is no continuation.
+    if (attr_parts.size() == 1)
+    {
+        string attr_str = NEW_LINE_INDENT + attr_name;
+        if (attr_value.codec_type == codec::codec_t::PERCENT)
+            attr_str += ATTRIBUTE_CONTINUATION_INDICATOR + codec::EQUAL_CHAR + attr_parts.at(0);
+        else
+            attr_str += codec::EQUAL_CHAR + codec::QUOTE_STR + attr_parts.at(0) + codec::QUOTE_STR;
+        return attr_str;
+    }
+
+    // All parts as continuations.
+
+    const vector<string>::size_type MAX_PARTS = 10;
+    vector<string>::size_type total_parts = std::min(MAX_PARTS, attr_parts.size());
+    vector<string>::size_type part_no = 0;
+    string attrs;
+    for (part_no = 0; part_no < total_parts; part_no++)
+    {
+        attrs += NEW_LINE_INDENT + attr_name + ATTRIBUTE_CONTINUATION_INDICATOR + to_string(part_no);
+        if (attr_value.codec_type == codec::codec_t::PERCENT)
+            attrs += ATTRIBUTE_CONTINUATION_INDICATOR + codec::EQUAL_CHAR + attr_parts.at(part_no);
+        else
+            attrs += codec::EQUAL_CHAR + codec::QUOTE_STR + attr_parts.at(part_no) + codec::QUOTE_STR;
+        if (part_no != total_parts - 1)
+            attrs += ATTRIBUTES_SEPARATOR_STR + codec::END_OF_LINE;
+    }
+
+    return attrs;
+}
+
+
 void mime::merge_attributes(attributes_t& attributes) const
 {
-    map<string, map<int, string>> attribute_parts;
+    map<string, map<int, string_t>> attribute_parts;
     for (auto attr = attributes.begin(); attr != attributes.end(); )
     {
         auto full_attr_name = attr->first;
         auto attr_value = attr->second;
-        string::size_type asterisk_pos = full_attr_name.find(ATTRIBUTE_MULTIPLE_NAME_INDICATOR);
+        string::size_type asterisk_pos = full_attr_name.find(ATTRIBUTE_CONTINUATION_INDICATOR);
+        // If no continuation indicator is found, then only one part is present, i.e. no continuation.
+        int attr_part = 1;
+        string attr_name = full_attr_name;
         if (asterisk_pos != string::npos)
         {
-            string attr_name = full_attr_name.substr(0, asterisk_pos);
             try
             {
-                if (!full_attr_name.substr(asterisk_pos + 1).empty())
-                {
-                    int attr_part = stoi(full_attr_name.substr(asterisk_pos + 1));
-                    attribute_parts[attr_name][attr_part] = attr_value;
-                    attr = attributes.erase(attr);
-                }
+                attr_name = full_attr_name.substr(0, asterisk_pos);
+                if (asterisk_pos < full_attr_name.length() - 1)
+                    attr_part = stoi(full_attr_name.substr(asterisk_pos + 1, 1));
             }
             catch (const std::invalid_argument& exc)
             {
-                throw mime_error("Parsing attribute failure at `" + attr_name + "`.");
+                throw mime_error("Parsing attribute name error.", "Attribute name is `" + attr_name + "`.");
             }
             catch (const std::out_of_range& exc)
             {
-                throw mime_error("Parsing attribute failure at `" + attr_name + "`.");
+                throw mime_error("Parsing attribute error.", "Attribute name is `" + attr_name + "`.");
             }
         }
-        else
-            attr++;
+        attribute_parts[attr_name][attr_part] = attr_value;
+        attr = attributes.erase(attr);
     }
 
     for (auto attr = attribute_parts.begin(); attr != attribute_parts.end(); attr++)
     {
         string attr_name = attr->first;
-        string attr_value;
+        string_t attr_value;
         for (auto part = attr->second.begin(); part != attr->second.end(); part++)
+        {
+            // Take the charset from the first attribute part, or set it to ASCII if not available.
+            // In case the other parts have the charset set, they are ignored.
+            if (attr_value.charset.empty())
+            {
+                if (!part->second.charset.empty())
+                    attr_value.charset = part->second.charset;
+                else
+                    attr_value.charset = codec::CHARSET_ASCII;
+            }
             attr_value += part->second;
-        attributes[attr_name] = attr_value;
+        }
+        attributes[attr_name] = decode_value_attribute(attr_value);
     }
 }
 
 
+string_t mime::decode_value_attribute(const string& attr_value) const
+{
+    if (attr_value.empty())
+        return string_t();
+
+    // percent decoding
+
+    size_t charset_pos = attr_value.find(codec::ATTRIBUTE_CHARSET_SEPARATOR);
+    if (charset_pos != string::npos)
+    {
+        size_t language_pos = attr_value.find(codec::ATTRIBUTE_CHARSET_SEPARATOR, charset_pos + 1);
+        if (language_pos == string::npos)
+            throw mime_error("Parsing attribute value error.", "No language parameter in the value `" + attr_value + "`.");
+
+        percent pct(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
+        return string_t(pct.decode(attr_value.substr(language_pos + 1)), attr_value.substr(0, charset_pos), codec::codec_t::PERCENT);
+    }
+
+    // Q decoding
+
+    q_codec qc(static_cast<string::size_type>(line_policy_), static_cast<string::size_type>(line_policy_));
+    auto av = qc.check_decode(attr_value);
+    return string_t(std::get<0>(av), std::get<1>(av), std::get<2>(av));
+}
+
+
 string mime::make_boundary() const
 {
     string bound;
-    bound.reserve(10);
+    static string::size_type BOUND_LEN = 10;
+    bound.reserve(BOUND_LEN);
     std::random_device rng;
     std::uniform_int_distribution<> index_dist(0, codec::HEX_DIGITS.size() - 1);
-    for (int i = 0; i < 10; i++)
+    for (string::size_type i = 0; i < BOUND_LEN; i++)
         bound += codec::HEX_DIGITS[index_dist(rng)];
     return BOUNDARY_DELIMITER + BOUNDARY_DELIMITER + BOUNDARY_DELIMITER + BOUNDARY_DELIMITER + bound;
 }
@@ -1117,7 +1294,7 @@ mime::media_type_t mime::mime_type_as_enum(const string& media_type_val) const
     else if (iequals(media_type_val, "message"))
         media_type = media_type_t::MESSAGE;
     else if (strict_mode_)
-        throw mime_error("Bad media type.");
+        throw mime_error("Bad media type.", "Given media type is `" + media_type_val + "`.");
     return media_type;
 }
 
